import { logger } from '@libp2p/logger';
import mergeOptions from 'merge-options';
// @ts-expect-error retimer does not have types
import retimer from 'retimer';
const log = logger('libp2p:connection-manager:auto-dialler');
const defaultOptions = {
    minConnections: 0,
    autoDialInterval: 10000
};
export class AutoDialler {
    /**
     * Proactively tries to connect to known peers stored in the PeerStore.
     * It will keep the number of connections below the upper limit and sort
     * the peers to connect based on wether we know their keys and protocols.
     */
    constructor(components, init) {
        this.components = components;
        this.options = mergeOptions.call({ ignoreUndefined: true }, defaultOptions, init);
        this.running = false;
        this._autoDial = this._autoDial.bind(this);
        log('options: %j', this.options);
    }
    isStarted() {
        return this.running;
    }
    /**
     * Starts the auto dialer
     */
    async start() {
        this.running = true;
        void this._autoDial().catch(err => {
            log.error('could not start autodial', err);
        });
        log('started');
    }
    /**
     * Stops the auto dialler
     */
    async stop() {
        this.running = false;
        if (this.autoDialTimeout != null) {
            this.autoDialTimeout.clear();
        }
        log('stopped');
    }
    async _autoDial() {
        if (this.autoDialTimeout != null) {
            this.autoDialTimeout.clear();
        }
        const minConnections = this.options.minConnections;
        // Already has enough connections
        if (this.components.connectionManager.getConnections().length >= minConnections) {
            this.autoDialTimeout = retimer(this._autoDial, this.options.autoDialInterval);
            return;
        }
        // Sort peers on whether we know protocols or public keys for them
        let peers = await this.components.peerStore.all();
        peers = peers.filter((peer) => {
            // do not dial ourselves
            if (peer.id.equals(this.components.peerId)) {
                return false;
            }
            // do not dial peers without multiaddrs
            if (peer.addresses.length === 0) {
                return false;
            }
            return true;
        });
        // shuffle the peers
        peers = peers.sort(() => Math.random() > 0.5 ? 1 : -1);
        peers = peers.sort((a, b) => {
            // dial peers with the most protocols first
            if (b.protocols.length > a.protocols.length) {
                return 1;
            }
            // dial peers with public keys first
            if (b.id.publicKey != null && a.id.publicKey == null) {
                return 1;
            }
            return -1;
        });
        for (let i = 0; this.running && i < peers.length && this.components.connectionManager.getConnections().length < minConnections; i++) {
            // Connection Manager was stopped during async dial
            if (!this.running) {
                return;
            }
            const peer = peers[i];
            if (this.components.connectionManager.getConnections(peer.id).length === 0) {
                log('connecting to a peerStore stored peer %p', peer.id);
                try {
                    await this.components.connectionManager.openConnection(peer.id);
                }
                catch (err) {
                    log.error('could not connect to peerStore stored peer', err);
                }
            }
        }
        // Connection Manager was stopped
        if (!this.running) {
            return;
        }
        this.autoDialTimeout = retimer(this._autoDial, this.options.autoDialInterval);
    }
}
//# sourceMappingURL=auto-dialler.js.map